@@ -1,9 +1,8 @@
 from __future__ import annotations
 
 import copy
-import inspect
 import string
-import typing
+from typing import Any, Callable, Generator, Iterable, List, Optional, Tuple, Union
 
 from .lazy import Lazy, resolve_lazy
 
@@ -15,6 +14,31 @@ EXCEPTION_HANDLER_NAME = "_htmlgenerator_exception_handler"
 "exception happens during rendering an element"
 
 
+def _render_element(
+    element: Any, context: dict, stringify: bool, fragment: Optional[str]
+) -> Generator[str, None, None]:
+    """Renders an element as a generator which yields strings
+    The stringify parameter should normally be true in order return
+    escaped strings. In some circumstances if is however desirable to
+    get the actual value and not a string returned. For such cases
+    stringify can be set to ``False``. An example are HTML attribute values
+    which us a ``hg.If`` element and return ``True`` or ``False`` to dynamically
+    control the appearance of the attribute.
+    The render_children method will use a default of ``True`` for strinfigy.
+    That behaviour should only be overriden by elements which consciously want
+    to be able to return non-string objects during rendering.
+    """
+    try:
+        while isinstance(element, Lazy):
+            element = element.resolve(context)
+        if isinstance(element, BaseElement):
+            yield from element.render(context, stringify=stringify, fragment=fragment)
+        elif element is not None and fragment is None:
+            yield conditional_escape(element) if stringify else element
+    except (Exception, RuntimeError) as e:
+        yield from _handle_exception(e, context)
+
+
 class BaseElement(list):
     """The base render element
     All nodes used in a render tree should have this class as a base.
@@ -27,64 +51,26 @@ class BaseElement(list):
         represents the child objects
         """
         super().__init__(children)
-        from . import DEBUG
-
-        if DEBUG:
-            # This will add the source location of where this element has
-            # been instantiated as a data attributte and a python attribute
-            # _src_location with (filename, linenumber, functionname) to this object
-            for frame in inspect.stack():
-                if frame.function != "__init__":
-                    break
-            if hasattr(self, "attributes"):
-                self.attributes[
-                    "data_source_location"
-                ] = f"{frame.filename}:{frame.lineno} in {frame.function}"
-            self._src_location = (frame.filename, frame.lineno, frame.function)
-
-    def _try_render(
-        self, element: typing.Any, context: dict, stringify: bool
-    ) -> typing.Generator[str, None, None]:
-        """Renders an element as a generator which yields strings
-        The stringify parameter should normally be true in order return
-        escaped strings. In some circumstances if is however desirable to
-        get the actual value and not a string returned. For such cases
-        stringify can be set to ``False``. An example are HTML attribute values
-        which us a ``hg.If`` element and return ``True`` or ``False`` to dynamically
-        control the appearance of the attribute.
-        The render_children method will use a default of ``True`` for strinfigy.
-        That behaviour should only be overriden by elements which consciously want
-        to be able to return non-string objects during rendering.
-        """
-        try:
-            while isinstance(element, Lazy):
-                element = element.resolve(context)
-            if isinstance(element, BaseElement):
-                yield from element.render(context)
-            elif element is not None:
-                yield conditional_escape(element) if stringify else element
-        except (Exception, RuntimeError) as e:
-            yield from _handle_exception(e, context)
 
     def render_children(
-        self, context: dict, stringify: bool = True
-    ) -> typing.Generator[str, None, None]:
+        self, context: dict, stringify: bool = True, fragment: Optional[str] = None
+    ) -> Generator[str, None, None]:
         """
         Renders all elements inside the list.
         Can be used by subclassing elements if they need to controll
         where child elements are rendered.
         """
         for element in self:
-            yield from self._try_render(element, context, stringify)
+            yield from _render_element(element, context, stringify, fragment)
 
     def render(
-        self, context: dict, stringify: bool = True
-    ) -> typing.Generator[str, None, None]:
+        self, context: dict, stringify: bool = True, fragment: Optional[str] = None
+    ) -> Generator[str, None, None]:
         """
         Renders this element and its children.
         Can be overwritten by subclassing elements.
         """
-        yield from self.render_children(context, stringify)
+        yield from self.render_children(context, stringify, fragment)
 
     """
     Tree functions
@@ -104,10 +90,8 @@ class BaseElement(list):
 
     def filter(
         self,
-        filter_func: typing.Callable[
-            [BaseElement, typing.Tuple[BaseElement, ...]], bool
-        ],
-    ) -> typing.Generator[BaseElement, None, None]:
+        filter_func: Callable[[BaseElement, Tuple[BaseElement, ...]], bool],
+    ) -> Generator[BaseElement, None, None]:
         """
         Walks through the tree (including self) and yields each
         element for which a call to filter_func evaluates to True.
@@ -119,9 +103,7 @@ class BaseElement(list):
 
     def wrap(
         self,
-        filter_func: typing.Callable[
-            [BaseElement, typing.Tuple[BaseElement, ...]], bool
-        ],
+        filter_func: Callable[[BaseElement, Tuple[BaseElement, ...]], bool],
         wrapperelement: BaseElement,
     ) -> list[BaseElement]:
         """
@@ -142,9 +124,7 @@ class BaseElement(list):
 
     def delete(
         self,
-        filter_func: typing.Callable[
-            [BaseElement, typing.Tuple[BaseElement, ...]], bool
-        ],
+        filter_func: Callable[[BaseElement, Tuple[BaseElement, ...]], bool],
     ) -> list[BaseElement]:
         """
         Walks through the tree (including self) and removes each element
@@ -164,7 +144,7 @@ class BaseElement(list):
 
     # untested code
     def _replace(
-        self, select_func: typing.Callable, replacement: BaseElement, all: bool = False
+        self, select_func: Callable, replacement: BaseElement, all: bool = False
     ):
         """Replaces an element which matches a certain condition with another element"""
         from .htmltags import HTMLElement
@@ -172,7 +152,7 @@ class BaseElement(list):
         class ReachFirstException(Exception):
             pass
 
-        def walk(element: typing.List, ancestors: typing.Tuple[BaseElement, ...]):
+        def walk(element: List, ancestors: Tuple[BaseElement, ...]):
             replacment_indices = []
             for i, e in enumerate(element):
                 if isinstance(e, BaseElement):
@@ -200,9 +180,9 @@ class BaseElement(list):
 class If(BaseElement):
     def __init__(
         self,
-        condition: typing.Union[bool, Lazy],
-        true_child: typing.Any,
-        false_child: typing.Any = None,
+        condition: Union[bool, Lazy],
+        true_child: Any,
+        false_child: Any = None,
     ):
         """
         condition: Value which determines which child to render
@@ -212,32 +192,36 @@ class If(BaseElement):
         super().__init__(true_child, false_child)
         self.condition = condition
 
-    def render(self, context: dict, stringify=True):
+    def render(
+        self, context: dict, stringify: bool = True, fragment: Optional[str] = None
+    ) -> Generator[str, None, None]:
         """The stringy argument can be set to False in order to get a python object
         instead of a rendered string returned. This is usefull when evaluating"""
         if resolve_lazy(self.condition, context):
-            yield from self._try_render(self[0], context, stringify)
+            yield from _render_element(self[0], context, stringify, fragment)
         elif len(self) > 1:
-            yield from self._try_render(self[1], context, stringify)
+            yield from _render_element(self[1], context, stringify, fragment)
 
 
 class Iterator(BaseElement):
     def __init__(
         self,
-        iterator: typing.Union[typing.Iterable, Lazy],
+        iterator: Union[Iterable, Lazy],
         loopvariable: str,
-        content: typing.Union[BaseElement, Lazy],
+        content: Union[BaseElement, Lazy],
     ):
         self.iterator = iterator
         self.loopvariable = loopvariable
         super().__init__(content)
 
-    def render(self, context: dict, stringify: bool = True):
+    def render(
+        self, context: dict, stringify: bool = True, fragment: Optional[str] = None
+    ) -> Generator[str, None, None]:
         context = dict(context)
         for i, value in enumerate(resolve_lazy(self.iterator, context)):
             context[self.loopvariable] = value
             context[self.loopvariable + "_index"] = i
-            yield from self.render_children(context, stringify)
+            yield from self.render_children(context, stringify, fragment)
 
 
 class WithContext(BaseElement):
@@ -256,20 +240,38 @@ class WithContext(BaseElement):
         self.additional_context = kwargs
         super().__init__(*children)
 
-    def render(self, context):
-        return super().render({**context, **self.additional_context})
+    def render(
+        self, context: dict, stringify: bool = True, fragment: Optional[str] = None
+    ) -> Generator[str, None, None]:
+        yield from super().render(
+            {**context, **self.additional_context},
+            stringify=stringify,
+            fragment=fragment,
+        )
+
+
+class Fragment(BaseElement):
+    def __init__(self, name, *children):
+        super().__init__(*children)
+        self.name = name
+
+    def render(
+        self, context: dict, stringify: bool = True, fragment: Optional[str] = None
+    ) -> Generator[str, None, None]:
+        if fragment is None or fragment == self.name:
+            yield from super().render(
+                context,
+                stringify=stringify,
+                fragment=None,  # must be None, render all subsequent elements
+            )
 
 
 def treewalk(
     element: BaseElement,
-    ancestors: typing.Tuple[BaseElement, ...],
-    filter_func: typing.Optional[
-        typing.Callable[[BaseElement, typing.Tuple[BaseElement, ...]], bool]
-    ],
-    apply: typing.Optional[
-        typing.Callable[[BaseElement, int, BaseElement], None]
-    ] = None,
-) -> typing.Generator[BaseElement, None, None]:
+    ancestors: Tuple[BaseElement, ...],
+    filter_func: Optional[Callable[[BaseElement, Tuple[BaseElement, ...]], bool]],
+    apply: Optional[Callable[[BaseElement, int, BaseElement], None]] = None,
+) -> Generator[BaseElement, None, None]:
     from .htmltags import HTMLElement
 
     matchelements = []
@@ -295,15 +297,17 @@ def treewalk(
             apply(element, i, e)
 
 
-def render(root: BaseElement, basecontext: dict) -> str:
+def render(root: BaseElement, basecontext: dict, fragment: Optional[str] = None) -> str:
     """Shortcut to serialize an object tree into a string"""
-    return mark_safe("").join(root.render(basecontext))
+    return mark_safe("").join(
+        root.render(basecontext, stringify=True, fragment=fragment)
+    )
 
 
 html_id_cache = set()
 
 
-def html_id(object: typing.Any, prefix: str = "id") -> str:
+def html_id(object: Any, prefix: str = "id") -> str:
     """Generate a unique HTML id from an object"""
     # Explanation of the chained call:
     # 1. id: We want a guaranteed unique ID. Since the lifespan of an HTML-response is
@@ -353,7 +357,8 @@ class ContextFormatter(string.Formatter):
         def extract(value):
             if isinstance(value, BaseElement):
                 return render(value, self.context)
-            return resolve_lazy(value, self.context) or ""
+            v = resolve_lazy(value, self.context)
+            return "" if v is None else v
 
         ret = super().get_value(
             key,
@@ -369,11 +374,14 @@ class FormatString(BaseElement):
         self.args = args
         self.kwargs = kwargs
 
-    def render(self, context):
-        yield from self._try_render(
+    def render(
+        self, context: dict, stringify: bool = True, fragment: Optional[str] = None
+    ) -> Generator[str, None, None]:
+        yield from _render_element(
             ContextFormatter(context).format(*self.args, **self.kwargs),
             context,
-            stringify=True,
+            stringify=True,  # must always be string
+            fragment=fragment,
         )
 
 
